#include  "axis-types.h"
#include  <cstdlib>
#include  <algorithm>
#include  <stdexcept>
#include  <cstring> 
#include  <cctype> 
#include  "lexer.h" 

namespace atlas { namespace interpreter {

const char* prim_names[]=
{"int","rat","string","bool",
     "vec", "mat", "ratvec",
     "LieType","RootDatum", "InnerClass", "RealForm",
     "CartanClass", "KGBElt", "Block", "Param", "Split", "ParamPol", 

 "void", nullptr };



const type_expr unknown_type; // uses default constructor
const type_expr void_type(empty_tuple());
const type_expr int_type(integral_type);
const type_expr bool_type(boolean_type);
const type_expr row_of_type(mk_type_expr("[*]"));
const type_expr gen_func_type(mk_type_expr("(*->*)"));

const type_expr rat_type(rational_type);
const type_expr str_type(string_type);
const type_expr vec_type(vector_type);
const type_expr ratvec_type(rational_vector_type);
const type_expr mat_type(matrix_type);
const type_expr row_of_int_type(mk_type_expr("[int]"));
const type_expr row_of_rat_type(mk_type_expr("[rat]"));
const type_expr row_of_vec_type(mk_type_expr("[vec]"));
const type_expr row_row_of_int_type(mk_type_expr("[[int]]"));
const type_expr pair_type(mk_type_expr("(*,*)"));
const type_expr int_int_type(mk_type_expr("(int,int)"));
const type_expr Lie_type_type(complex_lie_type_type);
const type_expr rd_type(root_datum_type);
const type_expr ic_type(inner_class_type);
const type_expr rf_type(real_form_type);
const type_expr split_type(split_integer_type);
const type_expr param_type(module_parameter_type);
const type_expr param_pol_type(virtual_module_type);

std::vector<shared_value> execution_stack;

std::vector<conversion_record> coerce_table;


type_ptr acquire(const type_expr* t) 
{ return type_ptr(new type_expr(t->copy())); }

type_list empty_tuple() { return type_list(); }

type_list& prefix(type_expr&& t, type_list& dst)
{ dst.push_front(std::move(t));
   return dst;
}
dressed_type_list& prefix(type_expr&& t, dressed_type_list& dst)
{ dst.push_front(std::move(t));
  return dst;
}

type_expr::type_expr(type_list&& l,bool is_union) noexcept
  : kind(is_union ? union_type: tuple_type)
  , tupple(l.release())
  {}

type_expr type_expr::copy() const
{ type_expr result;
  switch (result.kind=kind)
  { case undetermined_type: break;
    case primitive_type: result.prim=prim; break;
    case row_type:
      result.component_type=new type_expr(component_type->copy());
    break;
    case tuple_type: case union_type:
      
      {
        wtl_const_iterator it(tupple);
        type_list dst;
        for (type_list::iterator oit = dst.begin(); not it.at_end(); ++it,++oit)
          dst.insert(oit,it->copy());
        result.tupple = dst.release(); // incorporate and transfer ownership
      }
    break;
    case function_type: result.func=new func_type(func->copy()); break;
  }
  return result;
}

void type_expr::clear() noexcept
{ switch (kind)
  { case undetermined_type: case primitive_type: break;
    case row_type: delete component_type; break;
    case tuple_type: case union_type: delete tupple; break;
    case function_type: delete func; break;
  }
  kind = undetermined_type;
}

void type_expr::set_from(type_expr&& p) noexcept
{ assert (kind==undetermined_type);
   // logic should ensure this; we promised not to |throw|
  switch(kind=p.kind) // copy top node
  { case undetermined_type: break;
    case primitive_type: prim=p.prim; break;
    case function_type: func=p.func; break;
    case row_type: component_type=p.component_type; break;
    case tuple_type: case union_type: tupple = p.tupple; break;
  }
  p.kind=undetermined_type;
  // detach descendants, so |p.clear()| will destroy top-level only
}

type_expr::type_expr(type_expr&& x) noexcept // move constructor
: kind(x.kind)
{ switch(kind) // move top node
  { case undetermined_type: break;
    case primitive_type: prim=x.prim; break;
    case function_type: func=x.func; break;
    case row_type: component_type=x.component_type; break;
    case tuple_type: case union_type: tupple = x.tupple; break;
  }
  x.kind=undetermined_type;
  // detach descendants, so destructor of |x| will do nothing
}

type_expr& type_expr::operator=(type_expr&& x) noexcept // move assignment
{ if (this!=&x)
  { clear(); // detach anything previously linked
    set_from(std::move(x)); // move top level structure
  }
  return *this;
}

void type_expr::swap(type_expr& other) noexcept
{
  if (kind==other.kind) // an easy case
    switch(kind)
    { case undetermined_type: break; // no need to swap |nothing| fields
      case primitive_type: std::swap(prim,other.prim); break;
      case function_type: std::swap(func,other.func); break;
      case row_type: std::swap(component_type,other.component_type); break;
      case tuple_type: case union_type: std::swap(tupple,other.tupple); break;
    }
  else
  {
    type_expr t(std::move(other));
    other.set_from(std::move(*this));
    this->set_from(std::move(t));
  }
}

bool type_expr::specialise(const type_expr& pattern)
{ if (pattern.kind==undetermined_type)
    return true; // specialisation to \.* trivially succeeds.
  if (kind==undetermined_type) // specialising \.* also always succeeds,
    { set_from(pattern.copy()); return true; }
     // by setting |*this| to |pattern|
  if (pattern.kind!=kind) return false; // impossible to refine
  switch(kind)
  { case primitive_type: return prim==pattern.prim;
    case function_type:
      return func->arg_type.specialise(pattern.func->arg_type) 
         and func->result_type.specialise(pattern.func->result_type);
    case row_type: return component_type->specialise(*pattern.component_type);
    case tuple_type: case union_type:
     
     {
       wtl_iterator it0(tupple);
       wtl_const_iterator it1(pattern.tupple);
       while (not it0.at_end() and not it1.at_end() and it0->specialise(*it1))
         { ++it0; ++it1; }
       return it0.at_end() and it1.at_end();
       // whether both lists terminated
     }
    default: return true; // to keep the compiler happy, cannot be reached
  }
}

bool type_expr::can_specialise(const type_expr& pattern) const
{ if (pattern.kind==undetermined_type)
    return true; // specialisation to \.* trivially succeeds.
  if (kind==undetermined_type)
    return true;
  if (pattern.kind!=kind) return false; // impossible to refine
  switch(kind)
  { case primitive_type: return prim==pattern.prim;
    case function_type:
      return func->arg_type.can_specialise(pattern.func->arg_type) 
         and func->result_type.can_specialise(pattern.func->result_type);
    case row_type:
      return component_type->can_specialise(*pattern.component_type);
    case tuple_type: case union_type:
      
      {
        wtl_const_iterator it0(tupple);
        wtl_const_iterator it1(pattern.tupple);
        while (not it0.at_end() and not it1.at_end()
               and it0->can_specialise(*it1))
          { ++it0; ++it1; }
        return it0.at_end() and it1.at_end();
        // whether both lists terminated
      }
    default: return true; // to keep the compiler happy, cannot be reached
  }
}

void print(std::ostream& out, const raw_type_list& l,char sep)
{ wtl_const_iterator it(l);
  if (not it.at_end())
    while (out << *it, not (++it).at_end())
      out << sep;
}

std::ostream& operator<<(std::ostream& out, const func_type& f)
{
  out << '(';
  if (f.arg_type.kind==tuple_type or
      f.arg_type.kind==union_type and f.arg_type.tupple!=nullptr)
     print(out,f.arg_type.tupple,f.arg_type.kind==tuple_type?',':'|');
     // naked tuple or union
  else out << f.arg_type; // other component type
  out << "->";
  if (f.result_type.kind==tuple_type or
      f.result_type.kind==union_type and f.result_type.tupple!=nullptr)
     print(out,f.result_type.tupple,f.result_type.kind==tuple_type?',':'|');
     // naked tuple or union
  else out << f.result_type; // other component type
  out << ')';
  return out;
}

std::ostream& operator<<(std::ostream& out, const type_expr& t)
{ switch(t.kind)
  { case undetermined_type: out << '*'; break;
    case primitive_type: out << prim_names[t.prim]; break;
    case function_type: out << *t.func; break;
    case row_type: out << '[' << *t.component_type << ']'; break;
    case tuple_type:
      if (t.tupple==nullptr)
        out << "void";
      else
      {
         print(out << '(', t.tupple,',');
         out << ')';
      }
    break;
    case union_type:
      if (t.tupple==nullptr)
        out << "(*)";
      else
      {
         print(out << '(', t.tupple,'|');
         out << ')';
      }
    break;
  }
  return out;
}

bool operator== (const type_expr& x,const type_expr& y)
{ if (x.kind!=y.kind) return false;
  switch (x.kind)
  {  default:
// all cases are listed below, but compilers want a |default| to |return|
  case undetermined_type: return true;
    case primitive_type: return x.prim==y.prim;
    case function_type:
      return x.func->arg_type==y.func->arg_type
	 and x.func->result_type==y.func->result_type;
    case row_type: return *x.component_type==*y.component_type;
    case tuple_type: case union_type:
       
       {
         wtl_const_iterator it0(x.tupple);
         wtl_const_iterator it1(y.tupple);
         while (not it0.at_end() and not it1.at_end()
                and *it0==*it1)
           { ++it0; ++it1; }
         return it0.at_end() and it1.at_end();
         // whether both lists terminated
       }
  }
}

type_ptr mk_prim_type(primitive_tag p)
{ return p<nr_of_primitive_types ?
    type_ptr(new type_expr(p)) :
    type_ptr(mk_tuple_type(empty_tuple()));
}

type_ptr mk_function_type (type_expr&& a, type_expr&& r)
{ return type_ptr(new type_expr(std::move(a),std::move(r))); }

type_ptr mk_row_type(type_ptr&& c)
{ return type_ptr (new type_expr(std::move(c))); }

type_ptr mk_tuple_type (type_list&& l)
{ return type_ptr(new type_expr(std::move(l),false)); }

type_ptr mk_union_type (type_list&& l)
{ return type_ptr(l.singleton() ? new type_expr(std::move(l.front()))
                  : new type_expr(std::move(l),true));
}


type_p make_prim_type(unsigned int p)
{ return mk_prim_type(static_cast<primitive_tag>(p)).release(); }

type_p make_function_type(type_p a,type_p r)
{ return
    mk_function_type(std::move(*type_ptr(a)),std::move(*type_ptr(r))).release();
}

type_p make_row_type(type_p c) { return mk_row_type(type_ptr(c)).release(); }

type_p make_tuple_type(raw_type_list l)
{ type_list result(l); result.reverse();
    return mk_tuple_type(std::move(result)).release();
}

type_p make_union_type(raw_type_list l)
{ type_list result(l); result.reverse();
    return mk_union_type(std::move(result)).release();
}


raw_type_list make_type_singleton(type_p t)
{ type_list result;
  result.push_front(std::move(*type_ptr(t)));
  return result.release();
}

raw_type_list make_type_list(raw_type_list l,type_p t)
{ type_list tmp(l); // since |prefix| needs second argument an lvalue reference
  return prefix(std::move(*type_ptr(t)),tmp).release();
}

type_expr scan_in_parens(const char*& s);
type_expr scan_union_list(const char*& s);
type_expr scan_tuple_list(const char*& s);

type_expr scan_type(const char*& s)
{ if (*s=='(')
  { type_expr result=scan_in_parens(++s);
    if (*s++!=')')
      throw logic_error("Missing ')' in type");
    return result;
  }
  else if (*s=='[')
  {
    type_ptr p(new type_expr(scan_in_parens(++s)));
    if (*s++!=']')
      throw logic_error("Missing ']' in type");
    return type_expr(std::move(p));
  }
  else if (*s=='*') return ++s,type_expr(); // undetermined type
  else 
       { for (size_t i=0; i<nr_of_primitive_types; ++i)
         { const char* name=prim_names[i];
           size_t l= std::strlen(name);
           if (std::strncmp(s,name,l)==0 and not isalpha(s[l]))
           { s+=l;
             primitive_tag tag = static_cast<primitive_tag>(i);
             if (tag == nr_of_primitive_types) // then name scanned was |"void"|
               return type_expr(empty_tuple()); // which is equivalent to |"()"|
             return type_expr(tag); // otherwise this is a true primitive type
           }
         }
         throw logic_error("Type unrecognised");
       }
}

type_expr mk_type_expr(const char* s)
{ const char* orig=s;
  try
  { return scan_type(s); }
  catch (logic_error e)
  { std::cerr << e.what() << "; original string: '" << orig 
              << "' text remaining: '" << s << "'\n";
    throw;
  // make the error hard to ignore; if thrown probably aborts the program
  }
}

type_ptr mk_type(const char* s)
{ return type_ptr(new type_expr(mk_type_expr(s))); }
  

type_expr scan_in_parens(const char*& s)
{ type_expr a=scan_union_list(s);
  if (*s!='-' or s[1]!='>')
    return a;
  return type_expr(std::move(a),scan_union_list(s+=2));
    // construct function type
}

type_expr scan_union_list(const char*& s)
{ dressed_type_list variants;
  while (variants.emplace_back(scan_tuple_list(s)),*s=='|')
    ++s;
  return variants.size()==1
    ? std::move(variants.front())
    : type_expr(variants.undress(),true);
}

type_expr scan_tuple_list(const char*& s)
{ static const std::string term("|-)]");
  dressed_type_list members;
  if (term.find(*s)==std::string::npos)
    // only act on non-terminating characters
    while (members.emplace_back(scan_type(s)),*s==',')
      ++s;
  return members.size()==1
    ? std::move(members.front())
    : type_expr(members.undress(),false);
}

type_expr unknown_tuple(size_t n)
{ type_list tl;
  while (n-->0)
    prefix(type_expr(),tl);
  return type_expr(std::move(tl));
}

std::ostream& operator<< (std::ostream& out, const value_base& v)
{ v.print(out); return out; }

void row_value::print(std::ostream& out) const
{ out << '[';
  for (auto it=val.begin(); it!=val.end(); ++it)
    out << (it==val.begin() ? "" : ",") << **it;
   out << ']';
}

void tuple_value::print(std::ostream& out) const
{ out << '(';
  for (auto it=val.begin(); it!=val.end(); ++it)
    out << (it==val.begin() ? "" : ",") << **it;
   out << ')';
}

void push_tuple_components()
{ shared_tuple tuple=get<tuple_value>();
  for (size_t i=0; i<tuple->length(); ++i)
    push_value(tuple->val[i]); // push component
}

void wrap_tuple(size_t n)
{ std::shared_ptr<tuple_value> result = std::make_shared<tuple_value>(n);
  while (n-->0) // standard idiom; not |(--n>=0)|, since |n| is unsigned!
    result->val[n] =pop_value();
  push_value(std::move(result));
}

void union_value::print(std::ostream& out) const
{ out << *comp << '.' << main_hash_table->name_of(injector_name); }

shared_value& evaluation_context::elem(size_t i, size_t j)
{
  evaluation_context* p=this;
  while (i-->0 and (p=p->next.get())!=nullptr) {}
  assert(p!=nullptr and j<p->frame.size());
return p->frame[j];
}

void push_expanded(expression_base::level l, const shared_value& v)
{ if (l==expression_base::single_value)
    push_value(v);
  else if (l==expression_base::multi_value)
  { shared_tuple p = std::dynamic_pointer_cast<const tuple_value>(v);
    if (p==nullptr)
      push_value(v);
    else
      for (size_t i=0; i<p->length(); ++i)
        push_value(p->val[i]); // push components
  }
} 

void conversion::evaluate(level l) const
{ exp->eval();
  (*type.convert)();
  if (l==no_value)
    execution_stack.pop_back();
}

void conversion::print(std::ostream& out) const
{ out << type.name << ':' << *exp; }

void coercion(const type_expr& from,
              const type_expr& to,
              const char* s, conversion_info::conv_f f)
{ coerce_table.emplace_back(from,to,s,f); }

void voiding::evaluate(level l) const
{ exp->void_eval();
  if (l==single_value)
    wrap_tuple<0>();
}

void voiding::print(std::ostream& out) const
{ out << "voided:" << *exp; }

bool coerce(const type_expr& from_type, const type_expr& to_type,
	    expression_ptr& e)
{ if (to_type==void_type)
  {
     return true;
  } // syntactically voided here, |e| is unchanged
  for (auto it=coerce_table.begin(); it!=coerce_table.end(); ++it)
    if (from_type==*it->from and to_type==*it->to)
    { e.reset(new conversion(*it,std::move(e)));
      return true;
    }
  return false;
}

expression_ptr conform_types
(const type_expr& found, type_expr& required, expression_ptr&& d, const expr& e)
{ if (not required.specialise(found) and not coerce(found,required,d))
    throw type_error(e,found.copy(),std::move(required));
  return std::move(d);
}

const conversion_record* row_coercion(const type_expr& final_type,
                                            type_expr& component_type)
{ for (auto it=coerce_table.begin(); it!=coerce_table.end(); ++it)
    if (final_type==*it->to and it->from->kind==row_type)
      return component_type.specialise(*it->from->component_type)
        ? &*it
        : nullptr;
  return nullptr;
}

unsigned int is_close (const type_expr& x, const type_expr& y)
{ expression_ptr dummy(nullptr);
  if (x==y)
    return 0x7;
  if (x.kind==undetermined_type or y.kind==undetermined_type)
    return 0x0;
      // undetermined types do not specialise (or coerce), and are not close
  if (x.kind==primitive_type or y.kind==primitive_type)
  { unsigned int flags=0x0;
    if (coerce(x,y,dummy)) flags |= 0x1;
    if (coerce(y,x,dummy)) flags |= 0x2;
    return flags==0 ? flags : flags|0x4;
  }
  if (x.kind!=y.kind)
    return 0x0;
  if (x.kind==row_type)
    return is_close(*x.component_type,*y.component_type);
  if (x.kind!=tuple_type)
    return 0x0; // non-aggregate types are only close if equal
  unsigned int flags=0x7; // now we have two tuple types; compare components
  wtl_const_iterator it0(x.tupple);
  wtl_const_iterator it1(y.tupple);
  while (not it0.at_end() and not it1.at_end() 
         and (flags&=is_close(*it0,*it1))!=0)
  { ++it0; ++it1; }
  return it0.at_end() and it1.at_end() ? flags : 0x0;
}

bool broader_eq (const type_expr& a, const type_expr& b)
{ if (a==void_type or b==unknown_type)
    return true;
  if (a==unknown_type or b==void_type)
    return false;
  if (a.kind==primitive_type)
    return (is_close(a,b)&0x2)!=0; // whether |b| can be converted to |a|
  if (a.kind!=b.kind) // includes remaining cases where |b.kind==primitive_type|
    return false; // no broader between different kinds on non-primitive types
  if (a.kind==row_type)
    return broader_eq(*a.component_type,*b.component_type);
  if (a.kind==function_type)
    return a.func->arg_type==b.func->arg_type and 
    broader_eq(a.func->result_type,b.func->result_type);
  wtl_const_iterator itb(b.tupple);
  for (wtl_const_iterator ita(a.tupple);
       not ita.at_end(); ++ita,++itb)
    if (itb.at_end() or not broader_eq(*ita,*itb)) return false;
  return itb.at_end(); // if list of |a| has ended, that of |b| must as well
}

} }

